# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc


class ImmutableDict(collections_abc.Mapping):
    def __init__(self, somedict):
        self.dict = dict(somedict)   # make a copy
        self.hash = None

    def __getitem__(self, key):
        return self.dict[key]

    def __setitem__(self, key, value):
        modifiable = dict(self.dict)
        modifiable[key] = value
        return ImmutableDict(modifiable)

    def __repr__(self):
        return self.dict.__repr__()

    def __len__(self):
        return len(self.dict)

    def __iter__(self):
        return iter(self.dict)

    def __hash__(self):
        if self.hash is None:
            self.hash = hash(frozenset(self.dict.items()))
        return self.hash

    def __eq__(self, other):
        if isinstance(other, ImmutableDict):
            return self.dict == other.dict
        else:
            return self.dict == other



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import datetime
import uuid
from collections import deque

import ply.yacc
import pyrfc3339

from .edn_lex import tokens, lex, Keyword
from .exceptions import EDNDecodeError
from .immutable_dict import ImmutableDict
from .immutable_list import ImmutableList

from .compat import basestring, unicode

# Dummy statement to indicate that 'tokens' is used.
if tokens:
    pass

start = 'expressions'

_serializers = {}


class TaggedElement(object):
    def __str__(self):
        raise NotImplementedError("To be implemented by derived classes")


def add_tag(tag_name, tag_fn_or_class):
    """
    Add a custom tag handler.
    """
    assert isinstance(tag_name, basestring)
    _serializers[tag_name] = tag_fn_or_class


def remove_tag(tag_name):
    """
    Remove a custom tag handler.
    """
    assert isinstance(tag_name, basestring)
    del _serializers[tag_name]


def tag(tag_name):
    """
    Decorator for a function or a class to use as a tagged element.

    Example:

        @tag("dog")
        def parse_dog(name):
            return {
                "kind": "dog",
                "name": name,
                "message": "woof-woof",
            }

        parse("#dog \"Max\"")
        # => {"kind": "dog", "name": "Max", "message": "woof-woof"}
    """

    def _tag_decorator(fn_or_cls):
        _serializers[tag_name] = fn_or_cls
        return fn_or_cls
    return _tag_decorator


def p_term_leaf(p):
    """term : CHAR
            | STRING
            | HEX_INTEGER
            | INTEGER
            | FLOAT
            | KEYWORD
            | SYMBOL
            | RATIO
            | WHITESPACE"""
    p[0] = p[1]


def p_vector(p):
    """vector : VECTOR_START expressions VECTOR_END"""
    p[0] = ImmutableList(p[2])


def p_list(p):
    """list : LIST_START expressions LIST_END"""
    p[0] = tuple(p[2])


def p_set(p):
    """set : SET_START expressions MAP_OR_SET_END"""
    p[0] = frozenset(p[2])


def p_simple_map(p):
    """simple_map : MAP_START expressions MAP_OR_SET_END"""
    terms = p[2]
    if len(terms) % 2 != 0:
        raise EDNDecodeError('Even number of terms required for map')
    # partition terms in pairs
    p[0] = ImmutableDict((terms[i], terms[i+1]) for i in range(0, len(terms), 2))


def p_map(p):
    """map : simple_map"""
    p[0] = p[1]


def _apply_map_namespace_tag(ns, k):
    """
    Helper for the p_map_with_namespace_tag rule.
    Apply a namespace tag on a map key.
    """
    if not isinstance(k, Keyword):
        return k

    if k.namespace:
        # https://clojure.org/reference/reader#_maps
        if k.namespace == "_":
            return k.with_namespace(None)

        return k

    return k.with_namespace(ns)


def p_map_with_namespace_tag(p):
    """map : MAP_NAMESPACE_TAG simple_map"""
    ns = p[1]
    p[0] = ImmutableDict((
        (_apply_map_namespace_tag(ns, k), v)
        for k, v in p[2].items()))


def p_discarded_expressions(p):
    """discarded_expressions : DISCARD_TAG expression discarded_expressions
                             |"""
    p[0] = deque()


def p_expressions_expression_expressions(p):
    """expressions : expression expressions"""
    p[2].appendleft(p[1])
    p[0] = p[2]


def p_expressions_empty(p):
    """expressions : discarded_expressions"""
    p[0] = deque()


def p_expression(p):
    """expression : vector
                  | list
                  | set
                  | map
                  | term"""
    p[0] = p[1]


def p_expression_discard_expression_expression(p):
    """expression : DISCARD_TAG expression expression"""
    p[0] = p[3]


def p_expression_tagged_element(p):
    """expression : TAG expression"""
    tag = p[1]
    element = p[2]

    if tag == 'inst':
        length = len(element)
        hyphens_count = element.count('-')

        if length == 10 and hyphens_count == 2:
            output = datetime.datetime.strptime(element, '%Y-%m-%d').date()
        elif length == 7 and hyphens_count == 1:
            output = datetime.datetime.strptime(element, '%Y-%m').date()
        elif length == 4 and hyphens_count == 0:
            output = datetime.datetime.strptime(element, '%Y').date()
        else:
            output = pyrfc3339.parse(element)
    elif tag == 'uuid':
        output = uuid.UUID(element)
    elif tag in _serializers:
        output = _serializers[tag](element)
    else:
        raise NotImplementedError(
            u"Don't know how to handle tag ImmutableDict({})".format(tag))

    p[0] = output


def eof():
    raise EDNDecodeError('EOF Reached')


def p_error(p):
    if p is None:
        eof()
    else:
        raise EDNDecodeError(p)


def parse_all(text, input_encoding='utf-8', debug=False,
              write_ply_tables=True):
    """
    Parse all objects from the text and return a (possibly empty) list.

    ``debug`` and ``write_ply_tables`` arguments are passed to the Yacc parser.
    If ``debug`` is True, the parser writes a ``parser.out`` debugging file.
    If ``write_ply_tables`` is True, the parser writes grammar tables in a
    ``parsetab.py`` module on disk and reuse it in subsequent calls. If it
    can't write this cache file, it'll print a warning, and be a bit less
    efficient as it'll have to re-generate the tables every time. If you can't
    let it write its cache, you can disable this warning by passing
    `write_ply_tables=False`.
    """
    if not isinstance(text, unicode):
        text = text.decode(input_encoding)

    # See http://www.dabeaz.com/ply/ply.html#ply_nn36
    p = ply.yacc.yacc(debug=debug, write_tables=write_ply_tables)
    expressions = p.parse(text, lexer=lex())
    return list(expressions)


def parse(text, **kwargs):
    """
    Parse one object from the text. Return None if the text is empty.

    See parse_all for the accepted arguments.
    """
    expressions = parse_all(text, **kwargs)
    return expressions[0] if expressions else None



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import datetime
import decimal
import fractions
import re
import uuid

import pyrfc3339

from .immutable_dict import ImmutableDict
from .immutable_list import ImmutableList
from .char import Char
from .edn_lex import Keyword, Symbol
from .edn_parse import TaggedElement

from .compat import _PY3, long, basestring, unicode, unichr


DEFAULT_INPUT_ENCODING = 'utf-8'
DEFAULT_OUTPUT_ENCODING = 'utf-8'

# proper unicode escaping
# see http://stackoverflow.com/a/24519338
ESCAPE = re.compile(r'[\x00-\x1f\\"\b\f\n\r\t]', re.UNICODE)
ESCAPE_DCT = {
    '\\': '\\\\',
    '"': '\\"',
    '\b': '\\b',
    '\f': '\\f',
    '\n': '\\n',
    '\r': '\\r',
    '\t': '\\t',
}
SPECIAL_CHARS = {
    "\n": r"\newline",
    "\r": r"\return",
    " ": r"\space",
    "\t": r"\tab",
}

for __i in range(0x20):
    ESCAPE_DCT.setdefault(unichr(__i), '\\u{0:04x}'.format(__i))
del __i


def unicode_escape(string):
    """Return a edn representation of a Python string"""
    def replace(match):
        return ESCAPE_DCT[match.group(0)]
    return '"' + ESCAPE.sub(replace, string) + '"'


def dump_char(c):
    """Return an EDN representation of a Char."""
    # 33-126 = printable ASCII range, except space (code 32)
    if ord(c) in range(33, 127):
        return "\\{}".format(c)

    if c in SPECIAL_CHARS:
        return SPECIAL_CHARS[c]

    # [2:] to strip the '0x' prefix
    # str("0") because it must be an str and not a unicode in Python 2.
    return "\\u{}".format(hex(ord(c))[2:].upper().rjust(4, str("0")))


def indent_lines(lines, open_sym, close_sym, indent, indent_step):
    """
    Indents a data structure.

    ``lines`` is an array of strings composed of each element of the data
    structure. ``open_sym`` and ``close_sym`` are strings representing the
    opening and closing symbol of the data structure (i.e.: for a dict it
    would be '{' and '}', respectively). ``indent`` is an integer representing
    the number of spaces used to indent. ``indent_step`` is an integer
    representing the current level of indentation.
    """
    indent_prev = indent_step
    indent_step = indent_prev + indent
    # open symbol should not be indented
    result = [open_sym]

    indent_spaces = indent_step * ' '
    result += [indent_spaces + line for line in lines]

    # close symbol should be indented using the previous indentation level
    result += [indent_prev * ' ' + close_sym]

    return '\n'.join(result)


def seq(obj, **kwargs):
    return [udump(i, **kwargs) for i in obj]


def udump(obj,
          string_encoding=DEFAULT_INPUT_ENCODING,
          keyword_keys=False,
          sort_keys=False,
          sort_sets=False,
          indent=None,
          indent_step=0):
    """
    Dumps a formatted representation of a Python object.

    ``string_encoding`` (defaults to 'utf-8') is the encoding to be used if the
    object are bytes instead of strings. ``keyword_keys`` when True (defaults
    to False) converts the keys from a dict from string to keywords.
    ``sort_keys`` when True (defaults to False) sort dict keys alphabetically.
    ``sort_sets`` when True (defaults to False) sort sets alphabetically.
    ``indent`` when set to a positive integer (defaults to None) represents
    the number of spaces used to indent the object. ``indent_step`` (defaults
    to 0) represents the current indentation level when ``indent`` is different
    from None.
    """
    kwargs = {
        "string_encoding": string_encoding,
        "keyword_keys": keyword_keys,
        "sort_keys": sort_keys,
        "sort_sets": sort_sets,
        "indent": indent,
        "indent_step": indent_step + (indent or 0),
    }

    if obj is None:
        return 'nil'
    elif isinstance(obj, bool):
        return 'true' if obj else 'false'
    elif isinstance(obj, (int, long, float)):
        return unicode(obj)
    elif isinstance(obj, decimal.Decimal):
        return '{}M'.format(obj)
    elif isinstance(obj, (Keyword, Symbol)):
        return unicode(obj)
    elif isinstance(obj, Char):
        return dump_char(obj)
    # CAVEAT LECTOR! In Python 3 'basestring' is alised to 'str' above.
    # Furthermore, in Python 2 bytes is an instance of 'str'/'basestring' while
    # in Python 3 it is not.
    elif isinstance(obj, bytes):
        return unicode_escape(obj.decode(string_encoding))
    elif isinstance(obj, basestring):
        return unicode_escape(obj)
    elif isinstance(obj, tuple):
        lines = seq(obj, **kwargs)
        if indent is None:
            return '({})'.format(' '.join(lines))
        else:
            return indent_lines(lines, '(', ')', indent, indent_step)
    elif isinstance(obj, (list, ImmutableList)):
        lines = seq(obj, **kwargs)
        if indent is None:
            return '[{}]'.format(' '.join(lines))
        else:
            return indent_lines(lines, '[', ']', indent, indent_step)
    elif isinstance(obj, (set, frozenset)):
        if sort_sets:
            obj = sorted(obj)

        lines = seq(obj, **kwargs)
        if indent is None:
            return '#{{{}}}'.format(' '.join(lines))
        else:
            return indent_lines(lines, '#{', '}', indent, indent_step)
    elif isinstance(obj, (dict, ImmutableDict)):
        pairs = obj.items()

        if sort_keys:
            pairs = sorted(pairs, key=lambda p: str(p[0]))

        if keyword_keys:
            pairs = ((Keyword(k) if isinstance(k, (bytes, basestring)) else k, v) for k, v in pairs)

        lines = [' '.join(seq(p, **kwargs)) for p in pairs]
        if indent is None:
            return '{{{}}}'.format(' '.join(lines))
        else:
            return indent_lines(lines, '{', '}', indent, indent_step)
    elif isinstance(obj, fractions.Fraction):
        return '{}/{}'.format(obj.numerator, obj.denominator)
    elif isinstance(obj, datetime.datetime):
        return '#inst "{}"'.format(pyrfc3339.generate(obj, microseconds=True))
    elif isinstance(obj, datetime.date):
        return '#inst "{}"'.format(obj.isoformat())
    elif isinstance(obj, uuid.UUID):
        return '#uuid "{}"'.format(obj)
    elif isinstance(obj, TaggedElement):
        return unicode(obj)
    raise NotImplementedError(
        u"encountered object of type '{}' for which no known encoding is available: {}".format(
            type(obj), repr(obj)))


def dump(obj,
         string_encoding=DEFAULT_INPUT_ENCODING,
         output_encoding=DEFAULT_OUTPUT_ENCODING,
         keyword_keys=False,
         sort_keys=False,
         sort_sets=False,
         indent=None):
    outcome = udump(obj,
                    string_encoding=string_encoding,
                    keyword_keys=keyword_keys,
                    sort_keys=sort_keys,
                    sort_sets=sort_sets,
                    indent=indent)
    if _PY3:
        return outcome
    return outcome.encode(output_encoding)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

try:
    import collections.abc as collections_abc
except ImportError:
    import collections as collections_abc
import copy as _copy


class ImmutableList(collections_abc.Sequence, collections_abc.Hashable):
    def __init__(self, wrapped_list, copy=True):
        """Returns an immutable version of the given list. Optionally creates a shallow copy."""
        self._list = list(wrapped_list) if copy else wrapped_list
        self._hash = None

    def __repr__(self):
        return self._list.__repr__()

    def __eq__(self, other):
        if isinstance(other, ImmutableList):
            return self._list == other._list
        else:
            return self._list == other

    def _call_wrapped_list_method(self, method, *args):
        new_list = _copy.copy(self._list)
        getattr(new_list, method)(*args)
        return ImmutableList(new_list, copy=False)

    # collection.Sequence methods
    # https://docs.python.org/2/library/collections.html#collections-abstract-base-classes

    def __getitem__(self, index):
        return self._list[index]

    def __len__(self):
        return len(self._list)

    # collection.Hashable methods
    # https://docs.python.org/2/library/collections.html#collections-abstract-base-classes

    def __hash__(self):
        if self._hash is None:
            self._hash = hash(tuple(self._list))
        return self._hash

    # Other list methods https://docs.python.org/2/tutorial/datastructures.html#more-on-lists

    def insert(self, *args):
        return self._call_wrapped_list_method("insert", *args)

    def sort(self, *args):
        return self._call_wrapped_list_method("sort", *args)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from .edn_lex import Keyword, Symbol
from .edn_parse import parse as loads, parse_all as loads_all
from .edn_parse import add_tag, remove_tag, tag, TaggedElement
from .edn_dump import dump as dumps
from .exceptions import EDNDecodeError
from .immutable_dict import ImmutableDict
from .immutable_list import ImmutableList
from .char import Char

__all__ = (
    'ImmutableList',
    'ImmutableDict',
    'Keyword',
    'Symbol',
    'Char',
    'TaggedElement',
    'EDNDecodeError',
    'add_tag',
    'dumps',
    'loads',
    'loads_all',
    'remove_tag',
    'tag',
)



# -*- coding: utf-8 -*-
# TODO: Tests pass on Python 3.6, Disabled to not break tests on 2.7 :-(
# from __future__ import absolute_import, division, print_function, unicode_literals

from collections import OrderedDict
from uuid import uuid4, UUID
import random
import datetime
import fractions
import unittest

import pytz

from edn_format import edn_lex, edn_parse, \
    loads, dumps, Keyword, Symbol, ImmutableDict, ImmutableList, Char, \
    TaggedElement, add_tag, remove_tag, tag, \
    EDNDecodeError

from edn_format.compat import _PY3, unicode


class ConsoleTest(unittest.TestCase):
    def test_dumping(self):
        is_exception = False
        try:
            loads("[1 true nil]")
        except AttributeError:
            is_exception = True
        self.assertFalse(is_exception)


class KeywordTest(unittest.TestCase):
    def test_name(self):
        for name in ("foo", "a/b", "a/b/c", "a:::::::b"):
            self.assertEqual(name, Keyword(name).name)

    def test_namespace(self):
        self.assertIsNone(Keyword("foo").namespace)
        self.assertEqual("a", Keyword("a/b").namespace)
        self.assertEqual("a", Keyword("a/b/d").namespace)

    def test_with_namespace(self):
        for (expected, keyword, namespace) in (
            ("k", "k", None),
            ("k", "n/k", None),
            ("n/k", "k", "n"),
            ("n/k", "x/k", "n"),
            ("n/a/b/c", "x/a/b/c", "n"),
        ):
            self.assertEqual(Keyword(expected),
                             Keyword(keyword).with_namespace(namespace))


class EdnTest(unittest.TestCase):
    def check_lex(self, expected_output, actual_input):
        self.assertEqual(expected_output, str(list(edn_lex.lex(actual_input))))

    def test_lexer(self):
        self.check_lex("[LexToken(SYMBOL,None,1,0)]",
                       "nil")
        self.check_lex("[LexToken(SYMBOL,True,1,0)]",
                       "true")
        self.check_lex("[LexToken(INTEGER,123,1,0)]",
                       "123")
        self.check_lex("[LexToken(HEX_INTEGER,1,1,0)]",
                       "0x1")
        self.check_lex(
            "[LexToken(INTEGER,456,1,0), " +
            "LexToken(SYMBOL,None,1,4), " +
            "LexToken(SYMBOL,False,1,8)]",
            "456 nil false")
        self.check_lex("[LexToken(CHAR,'c',1,0)]" if _PY3 else "[LexToken(CHAR,u'c',1,0)]",
                       r"\c")
        self.check_lex("[LexToken(KEYWORD,Keyword(abc),1,0)]",
                       r":abc")
        self.check_lex("[LexToken(KEYWORD,Keyword(+),1,0)]",
                       r":+")
        self.check_lex("[]",
                       "; a comment")
        self.check_lex("[LexToken(KEYWORD,Keyword(abc),1,0)]",
                       ":abc ; a comment")
        self.check_lex("[LexToken(TAG,'inst',1,0), " +
                       "LexToken(STRING,'1985-04-12T23:20:50.52Z',1,6)]",
                       '#inst "1985-04-12T23:20:50.52Z"')
        self.check_lex("[LexToken(SYMBOL,Symbol(abc),1,0)]",
                       "abc")
        self.check_lex("[LexToken(SYMBOL,Symbol(?abc),1,0)]",
                       "?abc")
        self.check_lex("[LexToken(SYMBOL,Symbol(/),1,0)]",
                       "/")
        self.check_lex("[LexToken(SYMBOL,Symbol(prefix/name),1,0)]",
                       "prefix/name")
        self.check_lex("[LexToken(SYMBOL,Symbol(true.),1,0)]",
                       "true.")
        self.check_lex("[LexToken(SYMBOL,Symbol(nil.),1,0)]",
                       "nil.")
        self.check_lex("[LexToken(SYMBOL,Symbol($:ABC?),1,0)]",
                       "$:ABC?")
        self.check_lex("[LexToken(MAP_START,'{',1,0), "
                       "LexToken(KEYWORD,Keyword(a),1,2), "
                       "LexToken(SYMBOL,False,1,5), "
                       "LexToken(KEYWORD,Keyword(b),1,12), "
                       "LexToken(SYMBOL,False,1,15), "
                       "LexToken(MAP_OR_SET_END,'}',1,21)]",
                       "{ :a false, :b false }")

        self.check_lex("[LexToken(RATIO,Fraction(2, 3),1,0)]",
                       "2/3")
        self.check_lex("[LexToken(MAP_NAMESPACE_TAG,'a',1,0), "
                       "LexToken(MAP_START,'{',1,3), "
                       "LexToken(MAP_OR_SET_END,'}',1,4)]",
                       "#:a{}")

    def check_parse(self, expected_output, actual_input):
        self.assertEqual(expected_output, edn_parse.parse(actual_input),
                         actual_input)

    def check_parse_all(self, expected_output, actual_input):
        self.assertEqual(expected_output, edn_parse.parse_all(actual_input),
                         actual_input)

    def check_dumps(self, expected_output, actual_input, **kw):
        self.assertEqual(expected_output, dumps(actual_input, **kw))

    def test_parser_single_expressions(self):
        for expected, edn_string in (
            (1, "1"),
            (16768115, "0xFFDC73"),
            (Symbol("xFF"), "xFF"),
            (Symbol("a*b"), 'a*b'),
            ("ab", '"ab"'),
            ('a"b', r'"a\"b"'),
            ("blah\n", '"blah\n"'),
            ([1, 2, 3], "[1 2 3]"),
            ({1, 2, 3}, "#{1 2 3}"),
            ([1, True, None], "[1 true nil]"),
            (Char("c"), r"\c"),
            (Char("\n"), r"\newline"),
            (Char(u"Σ"), u"\\Σ"),
            (Char(u"λ"), r"\u03bB"),
            (Keyword("abc"), ":abc"),
            ([Keyword("abc"), 1, True, None], "[:abc 1 true nil]"),
            ((Keyword("abc"), 1, True, None), "(:abc 1 true nil)"),
            (tuple(), "()"),
            (set(), "#{}"),
            ({}, "{}"),
            ([], "[]"),
            ({"a": [1, 2, 3]}, '{"a" [1 2 3]}'),
            (datetime.datetime(2012, 12, 22, 19, 40, 18, 0, tzinfo=pytz.utc),
                '#inst "2012-12-22T19:40:18Z"'),
            (datetime.date(2011, 10, 9),
                '#inst "2011-10-09"'),
            ("|", "\"|\""),
            ("%", "\"%\""),
            (['bl"ah'], r"""["bl\"ah"]"""),
            ("blah\n", '"blah\n"'),
            ('"', r'"\""'),
            ('\\', r'"\\"'),
            (["abc", "123"], '["abc", "123"]'),
            ({"key": "value"}, '{"key" "value"}'),
            (frozenset({ImmutableList([u"ab", u"cd"]), ImmutableList([u"ef"])}),
                '#{["ab", "cd"], ["ef"]}'),
            (fractions.Fraction(2, 3), "2/3"),
            ((2, Symbol('/'), 3), "(2 / 3)"),

            ({}, "#:foo{}"),
            ({"a": "b"}, "#:foo{\"a\" \"b\"}"),
            ({Keyword("k"): 42}, "#:foo{:_/k 42}"),
            ({Keyword("bar/k"): 42}, "#:foo{:bar/k 42}"),
            ({Keyword("foo/k"): 42}, "#:foo{:k 42}"),
            ({Keyword("foo/k"): {Keyword("k"): 1}}, "#:foo{:k {:k 1}}"),
        ):
            self.check_parse(expected, edn_string)
            self.check_parse_all([expected], edn_string)

    def test_parser_multiple_expressions(self):
        for expected, edn_string in (
            ([], ""),
            ([], "              ,,,,          ,, ,     "),
            ([1], ",,,,,,,1,,,,,,,,,"),
            ([1, 2], "1 2"),
            ([0, Symbol("x1"), 1], "0 x1 0x1"),
            ([1, 2], "1                    2"),
            ([True, 42, False, Symbol('end')], "true 42 false end"),
            ([Symbol("a*b"), 42], 'a*b 42'),
            ([Char("a"), Char("b"), Char("c")], r"\a \b \c"),
        ):
            self.check_parse_all(expected, edn_string)
            if expected:
                self.check_parse(expected[0], edn_string)

    def check_roundtrip(self, data_input, **kw):
        self.assertEqual(data_input, loads(dumps(data_input, **kw)))

    def check_eof(self, data_input, **kw):
        with self.assertRaises(EDNDecodeError) as ctx:
            loads(data_input, **kw)

        self.assertEqual('EOF Reached', str(ctx.exception))

    def check_mismatched_delimiters(self):
        for bad_string in ("[", "(", "{", "(((((())", '"', '"\\"'):
            self.check_eof(bad_string)

    def test_dump(self):
        self.check_roundtrip({1, 2, 3})
        self.check_roundtrip({1, 2, 3}, sort_sets=True)
        self.check_roundtrip(
            {Keyword("a"): 1,
             "foo": Keyword("gone"),
             Keyword("bar"): [1, 2, 3]})
        self.check_roundtrip(uuid4())
        self.check_roundtrip(datetime.date(1354, 6, 7))
        self.check_roundtrip(datetime.datetime(1900, 6, 7, 23, 59, 59,
                                               tzinfo=pytz.utc))

    def test_proper_unicode_escape(self):
        self.check_roundtrip(u"\"")
        self.check_roundtrip(u'"')
        self.check_roundtrip(u"\\\"")
        self.check_roundtrip(u'\\"')
        self.check_roundtrip(u'\b\f\n\r\t"\\')

    def test_round_trip_conversion(self):
        edn_literals = [
            ["[ :ghi ]", "[:ghi]"],
            ["[:a #_foo 42]", "[:a 42]"],
            ["123N", "123"],
            ["-123N", "-123"],
            ["+123", "123"],
            ["+123N", "123"],
            ["123.2", "123.2"],
            ["+32.23M", "32.23M"],
            ["3.23e10", "32300000000.0"],
            ["3e10", "30000000000.0"],
        ]

        for literal in edn_literals:
            step1 = literal[0]
            step2 = loads(step1)
            step3 = dumps(step2)
            self.assertEqual(literal[1], step3)

    def test_round_trip_sets(self):
        step1 = '#{:a (1 2 3) :b}'
        step2 = loads(step1)
        step3 = dumps(step2)
        step4 = loads(step3)
        self.assertIn(Keyword("a"), step4)
        self.assertIn(Keyword("b"), step4)
        self.assertIn((1, 2, 3), step4)

    def test_round_trip_inst_short(self):
        step1 = '#inst "2011"'
        step2 = loads(step1)
        step3 = dumps(step2)
        step4 = loads(step3)
        self.assertEqual('#inst "2011-01-01"', step3)
        self.assertEqual(datetime.date(2011, 1, 1), step4)

    def test_round_trip_same(self):
        EDN_LITERALS = (
            "nil",
            "true",
            "false",
            '"hello world"',
            ":keyword",
            ":+",
            ":!",
            ":-",
            ":_",
            ":$",
            ":&",
            ":=",
            ":.",
            ":abc/def",
            "symbol",
            "123",
            "-123",
            "32.23",
            "32.23M",
            "-32.23M",
            "32M",
            "-32M",
            "-3.903495E-73M",
            "3.23e-10",
            "3e+20",
            "3E+20M",
            '["abc"]',
            '[1]',
            '[1 "abc"]',
            '[1 "abc" true]',
            '[:ghi]',
            '(:ghi)',
            '[1 "abc" true :ghi]',
            '(1 "abc" true :ghi)',
            '{"a" 2}',
            '#inst "1985-04-12T23:20:50.000000Z"',
            '#inst "2011-10-09"',
            '#uuid "f81d4fae-7dec-11d0-a765-00a0c91e6bf6"',
            '#date "19/07/1984"',
            '#{{"a" 1}}',
            '#{{"a" #{{:b 2}}}}',
            '"|"',
            "/",
            "1/3",
            '"1/3"',
            '"c"',
            r'"\n"',
            r"\c",
            r"\n",
            r"\newline",
            r"\tab",
            r"\uAA0A",
        )

        class TagDate(TaggedElement):
            def __init__(self, value):
                super(TagDate, self).__init__()
                self.name = 'date'
                self.value = datetime.datetime.strptime(
                    value,
                    "%d/%m/%Y").date()

            def __str__(self):
                return '#{} "{}"'.format(
                    self.name,
                    self.value.strftime("%d/%m/%Y"))

        add_tag('date', TagDate)

        for literal in EDN_LITERALS:
            step1 = literal
            step2 = loads(step1)
            step3 = dumps(step2)
            self.assertEqual(step1, step3)

    def check_char(self, expected, name):
        edn_data = "\\{}".format(name)
        parsed = loads(edn_data)
        self.assertIsInstance(parsed, unicode)
        self.assertEqual(expected, parsed, edn_data)

        self.assertIsInstance(parsed, Char)
        self.assertEqual(Char(expected), parsed, edn_data)

    def test_chars(self):
        # 33-126 = printable ASCII range, except space (code 32)
        for i in range(33, 127):
            ch = chr(i)
            self.check_char(ch, ch)

        for expected, name in (
                (" ", "space"),
                ("\n", "newline"),
                ("\r", "return"),
                ("\t", "tab"),
                (u"\uAA0A", "uAA0A"),
        ):
            self.check_char(expected, name)

    def test_exceptions(self):
        with self.assertRaises(EDNDecodeError):
            loads("{")

    def test_fractions(self):
        for edn_data in (
            '0/1',
            '1/1',
            '1/2',
            '1/3',
            '-5/3',
            '99999999999999999999999999999999999/999999999999999999999999991',
        ):
            self.assertEqual(edn_data, dumps(loads(edn_data)), edn_data)

    def test_keyword_keys(self):
        unchanged = (
            None,
            True,
            1,
            "foo",
            {},
            {True: 42},
            {25: 42},
            {3.14: "test"},
            {Keyword("foo"): "something"},
            {Symbol("foo"): "something"},
            ["foo", "bar"],
            ("foo", "bar"),
            {1: {2: 3}},
            ImmutableDict({1: 2}),
        )

        for case in unchanged:
            self.check_roundtrip(case, keyword_keys=True)

        keyworded_keys = (
            ("{:foo 42}", {"foo": 42}),
            ("{:a {:b {:c 42} :d 1}}", {"a": {"b": {"c": 42}, "d": 1}}),
            ("[{:a 42} {:b 25}]", [{"a": 42}, {"b": 25}]),
            ("({:a 42} {:b 25})", ({"a": 42}, {"b": 25})),
            ("{1 [{:a 42}]}", {1: [{"a": 42}]}),
            ("{:foo 1}", ImmutableDict({"foo": 1})),
        )

        for expected, data in keyworded_keys:
            self.assertEqual(expected, dumps(data, keyword_keys=True))

        self.assertEqual(
                {Keyword("a"): {Keyword("b"): 2}, 3: 4},
                loads(dumps({Keyword("a"): {"b": 2}, 3: 4}, keyword_keys=True)))

    def test_sort_keys(self):
        cases = (
            ('{"a" 4 "b" 5 "c" 3}', OrderedDict([("c", 3), ("b", 5), ("a", 4)])),
            ('{1 0 2 0 "a" 0}', {"a": 0, 1: 0, 2: 0}),
            ('[{"a" 1 "b" 1}]', [OrderedDict([("b", 1), ("a", 1)])]),
        )

        for expected, data in cases:
            self.check_dumps(expected, data, sort_keys=True)

        self.check_dumps('{:a 1 :b 1 :c 1 :d 1}',
                         {"a": 1, "d": 1, "b": 1, "c": 1},
                         sort_keys=True, keyword_keys=True)

    def test_sort_sets(self):
        def misordered_set_sequence():
            """"
            Return a tuple that, if put in a set then iterated over, doesn't
            yield elements in the original order
            """
            while True:
                seq = tuple(random.sample(range(10000), random.randint(2, 8)))
                s = set(seq)
                if tuple(s) != seq:
                    return seq

        for _ in range(10):
            seq = misordered_set_sequence()
            self.check_roundtrip(set(seq))
            self.check_dumps("#{{{}}}".format(" ".join(str(i) for i in sorted(seq))),
                             set(seq),
                             sort_sets=True)

    def test_indent(self):
        fixture = {"foo": "bar"}
        self.check_dumps('{"foo" "bar"}', fixture, indent=None)

        fixture = {Keyword("a"): {Keyword("b"): "foo"}}
        self.check_dumps('''\
{
:a {
:b "foo"
}
}''', fixture, indent=0)

        fixture = (1, Keyword("b"), "foo")
        self.check_dumps('''\
(
 1
 :b
 "foo"
)''', fixture, indent=1)

        fixture = [1, Keyword("b"), "foo", {Keyword("foo"): Keyword("bar")}]
        self.check_dumps('''\
[
  1
  :b
  "foo"
  {
    :foo :bar
  }
]''', fixture, indent=2)

        fixture = {1, 2, 3}
        self.check_dumps('''\
#{
    1
    2
    3
}''', fixture, indent=4, sort_sets=True)

        fixture = [{"a": 42}]
        self.check_dumps('''\
[
  {
    :a 42
  }
]''', fixture, keyword_keys=True, indent=2)

        fixture = [[1, 2, 3]]
        self.check_dumps('''\
[
  [
    1
    2
    3
  ]
]''', fixture, indent=2)

        fixture = {
            "a": "foo",
            "b": {"c": [Keyword("a"), Keyword("b"), Keyword("c")]},
            "d": {1, 2, 3},
            "e": (
                datetime.date(2011, 10, 9),
                UUID("urn:uuid:1e4856a2-085e-45df-93e1-41a0c7aeab2e"),
                datetime.datetime(2012, 12, 22, 19, 40, 18, 0, tzinfo=pytz.utc)
            )
        }
        self.check_dumps('''\
{
  :a "foo"
  :b {
    :c [
      :a
      :b
      :c
    ]
  }
  :d #{
    1
    2
    3
  }
  :e (
    #inst "2011-10-09"
    #uuid "1e4856a2-085e-45df-93e1-41a0c7aeab2e"
    #inst "2012-12-22T19:40:18.000000Z"
  )
}''', fixture, keyword_keys=True, sort_keys=True, sort_sets=True, indent=2)

    def test_discard(self):
        for expected, edn_data in (
            ('[x]', '[x #_ z]'),
            ('[z]', '[#_ x z]'),
            ('[x z]', '[x #_ y z]'),
            ('{1 4}', '{1 #_ 2 #_ 3 4}'),
            ('[1 2]', '[1 #_ [ #_ [ #_ [ #_ [ #_ 42 ] ] ] ] 2 ]'),
            ('[1 2 11]', '[1 2 #_ #_ #_ #_ 4 5 6 #_ 7 #_ #_ 8 9 10 11]'),
            ('()', '(#_(((((((1))))))))'),
            ('[6]', '[#_ #_ #_ #_ #_ 1 2 3 4 5 6]'),
            ('[4]', '[#_ #_ 1 #_ 2 3 4]'),
            ('{:a 1}', '{:a #_:b 1}'),
            ('[42]', '[42 #_ {:a [1 2 3 4] true false 1 #inst "2017"}]'),
            ('#{1}', '#{1 #_foo}'),
            ('"#_ foo"', '"#_ foo"'),
            (r'[\# _]', r'[\#_]'),
            ('[_]', r'[#_\#_]'),
            ('[1]', '[1 #_\n\n42]'),
            ('{}', '{#_ 1}'),
        ):
            self.assertEqual(expected, dumps(loads(edn_data)), edn_data)

    def test_discard_syntax_errors(self):
        for edn_data in ('#_', '#_ #_ 1', '#inst #_ 2017', '[#_]'):
            with self.assertRaises(EDNDecodeError):
                loads(edn_data)

    def test_discard_all(self):
        for edn_data in (
            '42', '-1', 'nil', 'true', 'false', '"foo"', '\\space', '\\a',
            ':foo', ':foo/bar', '[]', '{}', '#{}', '()', '(a)', '(a b)',
            '[a [[[b] c]] 2]', '#inst "2017"',
        ):
            self.assertEqual([1], loads('[1 #_ {}]'.format(edn_data)), edn_data)
            self.assertEqual([1], loads('[#_ {} 1]'.format(edn_data)), edn_data)

            self.assertEqual(None, loads('#_ {}'.format(edn_data)))

            for coll in ('[%s]', '(%s)', '{%s}', '#{%s}'):
                expected = coll % ""
                edn_data = coll % '#_ {}'.format(edn_data)
                self.assertEqual(expected, dumps(loads(edn_data)), edn_data)

    def test_chained_discards(self):
        for expected, edn_data in (
            ('[]', '[#_ 1 #_ 2 #_ 3]'),
            ('[]', '[#_ #_ 1 2 #_ 3]'),
            ('[]', '[#_ #_ #_ 1 2 3]'),
        ):
            self.assertEqual(expected, dumps(loads(edn_data)), edn_data)

    def test_custom_tags(self):
        @tag("dog")
        def parse_dog(name):
            return {
                "kind": "dog",
                "name": name,
                "message": "woof-woof",
            }

        @tag("cat")
        class Cat(TaggedElement):
            def __init__(self, name):
                self.name = name

        dog = loads("#dog \"Max\"")
        self.assertEqual(
                {"kind": "dog", "name": "Max", "message": "woof-woof"}, dog)

        cat = loads("#cat \"Alex\"")
        self.assertIsInstance(cat, Cat)
        self.assertEqual("Alex", cat.name)

        remove_tag("cat")
        self.assertRaises(NotImplementedError, lambda: loads("#cat \"Alex\""))

        add_tag("cat", Cat)
        cat = loads("#cat \"Alex\"")
        self.assertIsInstance(cat, Cat)
        self.assertEqual("Alex", cat.name)


class EdnInstanceTest(unittest.TestCase):
    def test_hashing(self):
        pop_count = len(set(map(hash,
                                ["db/id",
                                 Keyword("db/id"),
                                 Symbol("db/id")])))
        self.assertEqual(pop_count, 3)

    def test_equality(self):
        self.assertTrue("db/id" != Keyword("db/id"))
        self.assertTrue("db/id" != Symbol("db/id"))
        self.assertTrue(Symbol("db/id") != Keyword("db/id"))
        self.assertTrue("db/id" == "db/id")
        self.assertTrue(Keyword("db/id") == Keyword("db/id"))
        self.assertTrue(Symbol("db/id") == Symbol("db/id"))


class ImmutableListTest(unittest.TestCase):
    def test_list(self):
        x = ImmutableList([1, 2, 3])
        self.assertTrue(x == [1, 2, 3])

        self.assertTrue(x.index(1) == 0)
        self.assertTrue(x.count(3) == 1)
        self.assertTrue(x.insert(0, 0) == [0, 1, 2, 3])

        y = ImmutableList([3, 1, 4])
        self.assertTrue(y.sort() == [1, 3, 4])


class CharTest(unittest.TestCase):
    def test_new_ok(self):
        for c in " abc123\x12$*\n\r!\"{":
            self.assertEqual(c, Char(c), c)

    def test_new_fail(self):
        self.assertRaises(AssertionError, lambda: Char("some string"))


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
import sys

_PY3 = sys.version_info[0] == 3

# alias Python 2 types to their corresponding types in Python 3 if necessary
if _PY3:
    long = int
    basestring = str
    unicode = str
    unichr = chr
    def _bytes(s): return bytes(s, 'utf-8')
else:
    long = long
    basestring = basestring
    unicode = unicode
    unichr = unichr
    _bytes = bytes



try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

with open("requirements.txt") as f:
    requirements = f.read().strip().split("\n")


setup(name="edn_format_debug",
      version="0.1.0",
      author="DuneLeto",
      author_email="duneletogit@gmail.com",
      description="Swaroop's EDN format reader and writer in Python, w/ minor debug addition.",
      long_description=open('README.md').read(),
      long_description_content_type='text/markdown',
      url="https://github.com/DuneLeto/edn_format",
      install_requires=requirements,
      license="Apache 2.0",
      packages=['edn_format'],
      # http://pypi.python.org/pypi?%3Aaction=list_classifiers
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: Apache Software License',
          'Topic :: Software Development',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 3',
          'Operating System :: OS Independent',
      ])



# -*- coding: utf-8 -*-

from .compat import unicode


class Char(unicode):
    """
    This class represents a one-character string.
    """
    def __new__(cls, content):
        assert len(content) == 1, "Char can only contain one character."
        return super(Char, cls).__new__(cls, content)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

# proper unicode escaping
# see http://stackoverflow.com/a/24519338
import codecs
import decimal
import fractions
import logging
import re


import ply.lex

from .exceptions import EDNDecodeError
from .immutable_dict import ImmutableDict
from .char import Char

from .compat import unicode, _bytes


ESCAPE_SEQUENCE_RE = re.compile(r'''
    ( \\U........      # 8-digit hex escapes
    | \\u....          # 4-digit hex escapes
    | \\x..            # 2-digit hex escapes
    | \\[0-7]{1,3}     # Octal escapes
    | \\N\{[^}]+\}     # Unicode characters by name
    | \\[\\'"abfnrtv]  # Single-character escapes
    )''', re.UNICODE | re.VERBOSE)


def decode_escapes(s):
    def decode_match(match):
        return codecs.decode(match.group(0), 'unicode-escape')

    return ESCAPE_SEQUENCE_RE.sub(decode_match, s)


class BaseEdnType(object):
    def __init__(self, name):
        self._name = unicode(name)
        self._type = BaseEdnType

    @property
    def name(self):
        return self._name

    @property
    def type(self):
        return self._type

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self._name)

    def __hash__(self):
        return ImmutableDict(self.__dict__).__hash__()


class Keyword(BaseEdnType):
    def __init__(self, name):
        super(Keyword, self).__init__(name)
        self._type = Keyword

    def __str__(self):
        return ':{}'.format(self.name)

    @property
    def namespace(self):
        """
        Return the keyword's namespace. This may be ``None``.
        """
        if "/" in self.name:
            return self.name.split("/", 1)[0]

    def with_namespace(self, ns):
        """
        Return a new keyword with the given namespace. Use ``None`` or an empty
        string to remove it:

            >>> Keyword("foo/a").with_namespace("bar")
            Keyword("bar/a")

            >>> Keyword("a").with_namespace("bar")
            Keyword("bar/a")

            >>> Keyword("a/b").with_namespace(None)
            Keyword("b")
        """
        name = self.name
        if "/" in name:
            name = name.split("/", 1)[1]

        if not ns:
            return self.__class__(name)

        return self.__class__("%s/%s" % (ns, name))


class Symbol(BaseEdnType):
    def __init__(self, name):
        super(Symbol, self).__init__(name)
        self._type = Symbol

    def __str__(self):
        return self._name


# http://www.dabeaz.com/ply/ply.html
tokens = ('WHITESPACE',
          'CHAR',
          'STRING',
          'INTEGER',
          'HEX_INTEGER',
          'FLOAT',
          'RATIO',
          'SYMBOL',
          'KEYWORD',
          'VECTOR_START',
          'VECTOR_END',
          'LIST_START',
          'LIST_END',
          'MAP_START',
          'SET_START',
          'MAP_OR_SET_END',
          'TAG',
          'DISCARD_TAG',
          'MAP_NAMESPACE_TAG')

PARTS = {}
PARTS["non_nums"] = r"\w.*+!\-_?$%&=:#<>@"
PARTS["all"] = PARTS["non_nums"] + r"\d"
PARTS["first"] = r"\w*!_?$%&=<>@"
PARTS["special"] = r"\-+."
PARTS["start"] = \
    ("("
     "[{first}]"
     "|"
     "[{special}]"
     "[{non_nums}]"
     "|"
     "[{special}]"
     ")").format(**PARTS)
SYMBOL = ("("
          "{start}"
          "[{all}]*"
          r"\/"
          "[{all}]+"
          "|"
          r"\/"
          "|"
          "{start}"
          "[{all}]*"
          ")").format(**PARTS)
KEYWORD = (":"
           "("
           "[{all}]+"
           r"\/"
           "[{all}]+"
           "|"
           "[{all}]+"
           ")").format(**PARTS)
TAG = (r"\#"
       r"[a-zA-Z]"  # https://github.com/edn-format/edn/issues/30#issuecomment-8540641
       "("
       "[{all}]*"
       r"\/"
       "[{all}]+"
       "|"
       "[{all}]*"
       ")").format(**PARTS)

DISCARD_TAG = r"\#\_"
# https://clojure.org/reference/reader#_maps
MAP_NAMESPACE_TAG = "#:[{all}]+".format(**PARTS)

t_VECTOR_START = r'\['
t_VECTOR_END = r'\]'
t_LIST_START = r'\('
t_LIST_END = r'\)'
t_MAP_START = r'\{'
t_SET_START = r'\#\{'
t_MAP_OR_SET_END = r'\}'
t_ignore = ''.join([" ", "\t", "\n", ","])


def t_WHITESPACE(t):
    r"(\\newline)|(\\return)|(\\space)|(\\tab)"
    if t.value == r"\newline":
        t.value = Char("\n")
    elif t.value == r"\return":
        t.value = Char("\r")
    elif t.value == r"\space":
        t.value = Char(" ")
    elif t.value == r"\tab":
        t.value = Char("\t")

    return t


def t_CHAR(t):
    # uXXXX hex code or from "!" to "~" = all printable ASCII chars except the space
    # or unicode word chars
    r"(\\u[0-9A-Fa-f]{4}|\\[!-~\w])"
    t.value = Char(t.value[1]
                   if len(t.value) == 2
                   else _bytes(t.value).decode('raw_unicode_escape'))
    return t


def t_STRING(t):
    r'"([^"\\]*(\\.[^"\\]*)*)"'
    t.value = t.value[1:-1]
    t.value = decode_escapes(t.value)
    return t


def t_FLOAT(t):
    r"""[+-]?\d+(?:(?:\.\d+([eE][+-]?\d+)?|[eE][+-]?\d+)M?|M)"""
    e_value = 0
    if 'e' in t.value or 'E' in t.value:
        matches = re.search(r'[eE]([+-]?\d+)M?$', t.value)
        if matches is None:
            raise EDNDecodeError('Invalid float : {}'.format(t.value))
        e_value = int(matches.group(1))
    if t.value.endswith('M'):
        ctx = decimal.getcontext()
        t.value = decimal.Decimal(t.value[:-1]) * ctx.power(1, e_value)
    else:
        t.value = float(t.value) * pow(1, e_value)
    return t


def t_RATIO(t):
    r"""-?\d+/\d+"""
    numerator, denominator = t.value.split("/", 1)
    t.value = fractions.Fraction(int(numerator), int(denominator))
    return t


def t_INTEGER(t):
    # "No integer other than 0 may begin with 0."
    # https://github.com/edn-format/edn#integers
    r"""[+-]?(?:0(?!x)|[1-9]\d*)N?"""
    if t.value.endswith('N'):
        t.value = t.value[:-1]
    t.value = int(t.value)
    return t


def t_HEX_INTEGER(t):
    r"""[+-]?0x[A-F0-9]+"""
    t.value = int(t.value, 16)
    return t


def t_COMMENT(t):
    r'[;][^\n]*'
    pass  # ignore


@ply.lex.TOKEN(DISCARD_TAG)
def t_DISCARD_TAG(t):
    t.value = t.value[1:]
    return t


@ply.lex.TOKEN(TAG)
def t_TAG(t):
    t.value = t.value[1:]
    return t


@ply.lex.TOKEN(MAP_NAMESPACE_TAG)
def t_MAP_NAMESPACE_TAG(t):
    t.value = t.value[2:]
    return t


@ply.lex.TOKEN(KEYWORD)
def t_KEYWORD(t):
    t.value = Keyword(t.value[1:])
    return t


@ply.lex.TOKEN(SYMBOL)
def t_SYMBOL(t):
    if t.value == "nil":
        t.value = None
    elif t.value == "true":
        t.value = True
    elif t.value == "false":
        t.value = False
    else:
        t.value = Symbol(t.value)
    return t


def t_error(t):
    raise EDNDecodeError(
        "Illegal character '{c}' with lexpos {p} in the area of ...{a}...".format(
            c=t.value[0], p=t.lexpos, a=t.value[0:100]))


def lex(text=None, debug=False):
    kwargs = {}
    if debug:
        kwargs["debug"] = True
        kwargs["debuglog"] = logging.getLogger(__name__)
    lex = ply.lex.lex(reflags=re.UNICODE, **kwargs)
    if text is not None:
        lex.input(text)
    return lex
